#!/usr/bin/python3

# Copyright (c) 2020, 2021, 2022 Humanitarian OpenStreetMap Team
#
# This file is part of OSM-Fieldwork.
#
#     This is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     This is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with OSM-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import os
import logging
import sys
import json
from typing import Union

import mercantile
from shapely.geometry import shape
from shapely.ops import unary_union
from pySmartDL import SmartDL
from cpuinfo import get_cpu_info
import queue
import concurrent.futures
import threading
from osm_fieldwork.sqlite import DataFile, MapTile
from osm_fieldwork.xlsforms import xlsforms_path
from osm_fieldwork.yamlfile import YamlFile


log = logging.getLogger(__name__)


def dlthread(dest: str,
             mirrors: list,
             tiles: list,
             xy: bool,
             ):
    """
    Thread to handle downloads for Queue

    Args:
        dest (str): The filespec of the tile cache
        mirrors (list): The list of mirrors to get imagery
        tiles (list): The list of tiles to download
        xy (bool): Whether to swap the X & Y fields in the TMS URL
    """
    if len(tiles) == 0:
        # epdb.st()
        return
    # counter = -1

    # start = datetime.now()

    # totaltime = 0.0
    log.info(
        "Downloading %d tiles in thread %d to %s"
        % (len(tiles), threading.get_ident(), dest)
    )
    for tile in tiles:
        bingkey = mercantile.quadkey(tile)
        filespec = f"{tile[2]}/{tile[1]}/{tile[0]}"
        for site in mirrors:
            if site["source"] != "topo":
                filespec += "." + site["suffix"]
            url = site["url"]
            if  site["source"] == "bing":
                remote = url % bingkey
            elif  site["source"] == "google":
                path = f"x={tile[0]}&s=&y={tile[1]}&z={tile[2]}"
                remote = url % path
            elif  site["source"] == "custom":
                if not xy:
                    path = f"x={tile[0]}&s=&y={tile[1]}&z={tile[2]}"
                else:
                    path = f"x={tile[0]}&s=&y={tile[2]}&z={tile[1]}"
                remote = url % path
            else:
                remote = url % filespec
            print("Getting file from: %s" % remote)
            # Create the subdirectories as pySmartDL doesn't do it for us
            if os.path.isdir(dest) is False:
                tmp = ""
                paths = dest.split("/")
                for i in paths[1:]:
                    tmp += "/" + i
                    if os.path.isdir(tmp):
                        continue
                    else:
                        os.mkdir(tmp)
                        log.debug("Made %s" % tmp)

        try:
            if site["source"] == "topo":
                filespec += "." + site["suffix"]
            outfile = dest + "/" + filespec
            if not os.path.exists(outfile):
                dl = SmartDL(remote, dest=outfile, connect_default_logger=False)
                dl.start()
            else:
                log.debug("%s exists!" % (outfile))
        except Exception as e:
            log.error(e)
            log.error("Couldn't download from %r: %s" % (filespec, dl.get_errors()))


class BaseMapper(object):
    def __init__(self,
                 boundary: str,
                 base: str,
                 source: str,
                 xy: bool,
                 ):
        """
        Create an mbtiles basemap for ODK Collect

        Args:
            boundary (str): A BBOX string or GeoJSON file of the AOI.
                The GeoJSON can contain multiple geometries.
            base (str): The base directory to cache map tile in
            source (str): The upstream data source for map tiles
            xy (bool): Whether to swap the X & Y fields in the TMS URL

        Returns:
            (BaseMapper): An instance of this class
        """
        self.bbox = self.makeBbox(boundary)
        self.tiles = list()
        self.base = base
        # sources for imagery
        self.source = source
        self.sources = dict()
        self.xy = xy

        path = xlsforms_path.replace("xlsforms", "imagery.yaml")
        self.yaml = YamlFile(path)

        for entry in self.yaml.yaml['sources']:
            for k, v in entry.items():
                src = dict()
                for item in v:
                    src['source'] = k
                    for k1, v1 in item.items():
                        # print(f"\tFIXME2: {k1} - {v1}")
                        src[k1] = v1
                self.sources[k] = src

    def customTMS(self,
                  url: str,
                  name: str = 'custom',
                  source: str = 'custom',
                  suffix: str = 'jpg',
                  ):
        """
        Add a custom TMS URL to the list of sources.

        Args:
            name (str): The name to display
            url (str): The URL string
            suffix (str): The suffix, png or jpg
            source (str): The source value to use as an index
        """
        self.sources['custom'] = {'name': name, 'url': url, 'suffix': suffix, 'source': source}
        self.source = 'custom'

    def getFormat(self):
        """
        Returns:
            (str): the upstream source for map tiles
        """
        return self.sources[self.source]["suffix"]

    def getTiles(self,
                 zoom: int = None,
                 ):
        """
        Get a list of tiles for the specifed zoom level

        Args:
            zoom (int): The Zoom level of the desired map tiles

        Returns:
            (int): The total number of map tiles downloaded
        """
        info = get_cpu_info()
        cores = info["count"]

        self.tiles = list(
            mercantile.tiles(
                self.bbox[0], self.bbox[1], self.bbox[2], self.bbox[3], zoom)
        )
        total = len(self.tiles)
        log.info("%d tiles for zoom level %d" % (len(self.tiles), zoom))
        chunk = round(len(self.tiles) / cores)
        queue.Queue(maxsize=cores)
        log.info("%d threads, %d tiles" % (cores, total))

        mirrors = [self.sources[self.source]]
        # epdb.st()
        if len(self.tiles) < chunk or chunk == 0:
            dlthread(self.base, mirrors, self.tiles, self.xy)
        else:
            with concurrent.futures.ThreadPoolExecutor(max_workers=cores) as executor:
                block = 0
                while block <= len(self.tiles):
                    executor.submit(
                        dlthread, self.base, mirrors, self.tiles[block : block + chunk]
                    )
                    log.debug("Dispatching Block %d:%d" % (block, block + chunk))
                    block += chunk
                executor.shutdown()
            # log.info("Had %r errors downloading %d tiles for data for %r" % (self.errors, len(tiles), os.path.basename(self.base)))

        return len(self.tiles)

    def tileExists(self,
                   tile: MapTile,
                   ):
        """
        See if a map tile already exists

        Args:
            tile (MapTile): The map tile to check for the existence of

        Returns:
            (bool): Whether the tile exists in the map tile cache
        """
        filespec = f"{self.base}{tile[2]}/{tile[1]}/{tile[0]}.{self.sources[{self.source}]['suffix']}"
        if os.path.exists(filespec):
            log.debug("%s exists" % filespec)
            return True
        else:
            log.debug("%s doesn't exists" % filespec)
            return False

    def makeBbox(self,
                 boundary: str,
                 ):
        """
        Make a bounding box from a shapely geometry.

        Args:
            boundary (str): A BBOX string or GeoJSON file of the AOI.
                The GeoJSON can contain multiple geometries.

        Returns:
            (list): The bounding box coordinates
        """
        if not boundary.lower().endswith((".json", ".geojson")):
            # Is BBOX string
            try:
                bbox_parts = boundary.split(',')
                bbox = tuple(float(x) for x in bbox_parts)
                if len(bbox) == 4:
                    # BBOX valid
                    return bbox
                else:
                    log.error(f"BBOX string malformed: {bbox}")
                    return
            except Exception as e:
                log.error(e)
                log.error(f"Failed to parse BBOX string: {boundary}")
                return

        log.debug(f"Reading geojson file: {boundary}")
        with open(boundary, "r") as f:
            poly = json.load(f)
        if 'features' in poly:
            geometry = shape(poly['features'][0]['geometry'])
        else:
            geometry = shape(poly)

        if isinstance(geometry, list):
            # Multiple geometries
            log.debug("Creating union of multiple bbox geoms")
            geometry = unary_union(geometry)

        if geometry.is_empty:
            log.warning(f"No bbox extracted from {geometry}")
            return None

        bbox = geometry.bounds
        # left, bottom, right, top
        # minX, minY, maxX, maxY
        return bbox


def main():
    """This main function lets this class be run standalone by a bash script"""
    parser = argparse.ArgumentParser(
        description="Create an mbtiles basemap for ODK Collect"
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("-b", "--boundary", required=True, help="The boundary for the area you want")
    parser.add_argument("-t", "--tms", help="Custom TMS URL")
    parser.add_argument("--xy", default=False, help="Swap the X & Y coordinates when using a custom TMS")
    parser.add_argument("-o", "--outfile", required=True, help="Output file name")
    parser.add_argument("-z", "--zooms", default="12-17", help="The Zoom levels")
    parser.add_argument("-d", "--outdir", help="Output directory name for tile cache")
    parser.add_argument("-s", "--source", default="esri",
        choices=["esri", "bing", "topo", "google", "oam"],
        help="Imagery source",
    )
    args = parser.parse_args()

    # if verbose, dump to the terminal.
    if args.verbose is not None:
        log.setLevel(logging.DEBUG)
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(threadName)10s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        log.addHandler(ch)

    # Get all the zoom levels we want
    zooms = list()
    if args.zooms:
        if args.zooms.find("-") > 0:
            start = int(args.zooms.split("-")[0])
            end = int(args.zooms.split("-")[1]) + 1
            x = range(start, end)
            for i in x:
                zooms.append(i)
        elif args.zooms.find(",") > 0:
            levels = args.zooms.split(",")
            for level in levels:
                zooms.append(int(level))
        else:
            zooms.append(int(args.zooms))

    # Make a bounding box from the boundary file
    if not args.boundary:
        log.error("You need to specify a boundary file!")
        parser.print_help()
        quit()

    if not args.outdir:
        base = "/var/www/html"
    else:
        base = args.outdir
    base = f"{base}/{args.source}tiles"

    if args.source and not args.tms:
        basemap = BaseMapper(args.boundary, base, args.source, args.xy)
    elif args.tms:
        basemap = BaseMapper(args.boundary, base, 'custom', args.xy)
        basemap.customTMS(args.tms)
    else:
        log.error("You need to specify a source!")
        parser.print_help()
        quit()

    if args.outfile is None:
        log.error("You need to specify an mbtiles or sqlitedb file!!")
        parser.print_help()
        quit()

    outf = DataFile(args.outfile, basemap.getFormat())
    suffix = os.path.splitext(args.outfile)[1]
    if suffix == ".mbtiles":
        outf.addBounds(basemap.bbox)
    for level in zooms:
        basemap.getTiles(level)
        if args.outfile:
            # Create output database and specify image format, png, jpg, or tif
            outf.writeTiles(basemap.tiles, base)
        else:
            log.info("Only downloading tiles to %s!" % base)

if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of osm_fieldwork.
#
#     Underpass is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with osm_fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import sys
import os
from osm_fieldwork.xlsforms import xlsforms_path
from osm_fieldwork.yamlfile import YamlFile

parser = argparse.ArgumentParser(description="Read and parse a YAML file")
parser.add_argument(
    "--infile", help="The YAML input file"
)
args = parser.parse_args()

path = xlsforms_path.replace("/xlsforms", "")
infile = f"{path}/xforms.yaml"
data = YamlFile(infile)
# data.dump()

def test_load():
    """See if the file got loaded"""
    hits = 0
    if len(data.yaml.keys()) > 0:
        hits = 1
    if len(data.yaml['convert']) > 0:
        hits += 1
    if len(data.yaml['ignore']) > 0:
        hits += 1
    if len(data.yaml['private']) > 0:
        hits += 1
    assert hits == 4

def test_good():
    hits = 0
    if data.convertData('amenity'):
        hits += 1
    if data.convertData('foobar') == False:
        hits += 1
    assert hits == 2

def test_private():
    hits = 0
    if data.privateData('income'):
        hits += 1
    if data.privateData('foobar') == False:
        hits += 1
    assert hits == 2

def test_ignore():
    hits = 0
    if data.ignoreData('model'):
        hits += 1
    if data.ignoreData('foobar') == False:
        hits += 1
    assert hits == 2

# def test_bool_bad():
#     assert "bad keyword" data.yaml['convert'] is not True


# def test_value():
#     assert "caravans" == "tourism"


if __name__ == "__main__":
    test_load()
    test_good()
    test_private()
    test_ignore()



#!/usr/bin/python3

# Copyright (c) 2023 Humanitarian OpenStreetMap Team
#
# This file is part of osm_fieldwork.
#
#     This is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with osm_fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import os
import sys
from osm_fieldwork.odk_merge import OdkMerge, conflateThread
from osm_fieldwork.osmfile import OsmFile

# find the path of root tests dir
rootdir = os.path.dirname(os.path.abspath(__file__))

parser = argparse.ArgumentParser(description="Test odk_merge")
parser.add_argument("--odk", default=f"{rootdir}/testdata/odk_pois.osm", help="The ODK file")
parser.add_argument("--osm", default=f"{rootdir}/testdata/osm_buildings.geojson", help="The OSM data")
parser.add_argument("-d", "--database", default=f"PG:colorado", help="The database name")
parser.add_argument("-b", "--boundary", default=f"{rootdir}/testdata/Salida.geojson", help="The project AOI")
args = parser.parse_args()

def test_file():
    """This tests conflating against the GeoJson data extract file"""
    passes = 0
    osm = OsmFile()
    osmdata = osm.loadFile(args.odk)
    odk = OdkMerge(args.osm)
    # Although the code is multi-threaded, we can call the function that
    # does all the work directly without threading. Easier to debug this qay.
    data = conflateThread(osmdata, odk, 0)
    # There are 8 features in the test data
    if len(data) == 8:
        passes += 1
    # The first feature is a match, so has the OSM ID, the second
    # feature doesn't match, so negative ID
    if data[0]['attrs']['id'] > 0 and data[1]['attrs']['id'] < 0:
        passes += 1
    # duplicates have a fixme tag added
    if 'fixme' in data[0]['tags'] and 'fixme' not in data[1]['tags']:
        passes += 1
    assert(passes == 3)

# FIXME update test_db to use local db in CI
# def test_db():
#     """This test against a local database. If there is no postgres, then
#     none of the tests get run."""
#     passes = 0
#     # this database always exists on this developer's machine
#     odk = OdkMerge("PG:colorado", args.boundary)
#     if odk.postgres is not None:
#         passes += 1
#     else:
#         return
#     # We also want to trap a bad database name
#     # odk = OdkMerge("PG:foobar")
#     # if odk.dbcursor is None:
#     #     passes += 1

#     osm = OsmFile()
#     osmdata = osm.loadFile(args.odk)
#     if len(osmdata) == 8:
#         passes += 1

#     # The first feature is a match, so has the OSM ID, the second
#     # feature doesn't match, so negative ID. Here we're just making
#     # sure it got loaded as a sanity check
#     keys = list(osmdata.keys())
#     if osmdata[keys[6]]['attrs']['id'] > 0 and osmdata[keys[0]]['attrs']['id'] < 0:
#         passes += 1

#     odk = OdkMerge(args.database, args.boundary)
#     # Although the code is multi-threaded, we can call the function that
#     # does all the work directly without threading. Easier to debug this way.
#     data = conflateThread(osmdata, odk, 0)
#     if len(data[6]['attrs']) > 0 and data[6]['attrs']['id'] > 0 and data[0]['attrs']['id'] < 0:
#         passes += 1
#     assert(passes == 4)

if __name__ == "__main__":
    print("--- test_file() ---")
    test_file()
    # print("--- test_db() ---")
    # test_db()
    print("--- done ---")



#!/usr/bin/python3

# Copyright (c) 2023 Humanitarian OpenStreetMap Team
#
# This file is part of osm_fieldwork.
#
#     This is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with osm_fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import os
import sys
from osm_fieldwork.make_data_extract import uriParser

parser = argparse.ArgumentParser(
    description="Test odk_merge"
)
parser.add_argument("--infile", help="The input file")
args = parser.parse_args()

out = None
def test_dbname_only():
    dbname = "testdb"
    db = uriParser(dbname)
    # print(db['dbname'] == dbname)
    assert(db['dbname'] == dbname)

def test_dbname():
    passes = 0
    db = uriParser("fmtm:XxXx@localhost/underpass")
    if db['dbname'] == 'underpass':
        passes += 1
    # print(db)
    db = uriParser("fmtm:XxXx/underpass")
    if db['dbname'] == 'underpass':
        passes += 1
    # print(db)
    db = uriParser("fmtm:XxXx@localhost:5433/underpass")
    if db['dbname'] == 'underpass':
        passes += 1
    #print(db)
    # print(passes == 3)
    assert(passes == 3)
    
def test_host():
    db = uriParser("fmtm@localhost")
    dbhost = db['dbhost']
    db2 = uriParser("fmtm@hostlocal")
    db2host = db2['dbhost']
    # print(dbhost == 'localhost' and db2host == 'hostlocal')
    assert(dbhost == 'localhost' and db2host == 'hostlocal')

def test_user():
    db = uriParser("fmtm@localhost")
    dbuser = db['dbuser']
    dbpass = db['dbpass']
    # print(dbuser == 'fmtm', dbpass is None)
    assert(dbuser == 'fmtm' and dbpass is None)

def test_password():
    db = uriParser("fmtm:XxXx@localhost")
    dbuser = db['dbuser']
    dbpass = db['dbpass']
    # print(dbuser == 'fmtm', dbpass == 'XxXx')
    assert(dbuser == 'fmtm' and dbpass == 'XxXx')

def test_port():
    db2 = uriParser("fmtm:XxXx@localhost:5433")
    # print(db2)
    db3 = uriParser("fmtm:XxXx@hostlocal:5433/fmtmdb")
    # print(db3)
    # print(db2['dbport'] == '5433' and db3['dbport'] == '5433')
    assert(db2['dbport'] == '5433' and db3['dbport'] == '5433')

if __name__ == "__main__":
    # print("--- test_dbname_only() ---")
    test_dbname_only()
    # print("--- test_dbname() ---")
    test_dbname()
    # print("--- test_host() ---")
    test_host()
    # print("--- test_user() ---")
    test_user()
    # print("--- test_password() ---")
    test_password()
    # print("--- test_port() ---")
    test_port()
    # print("--- done ---")



#!/usr/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of osm-Fieldwork.
#
#     Osm-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with osm_fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import argparse
import logging
import sys
import os
import pandas as pd
from geojson import Feature, FeatureCollection
import geojson
from osm_fieldwork.xlsforms import xlsforms_path
import yaml

# Find the other files for this project
import osm_fieldwork as of
rootdir = of.__path__[0]

# Instantiate logger
log = logging.getLogger(__name__)


class FilterData(object):
    def __init__(self,
                 filespec: str = None
                 ):
        """

        Args:
            filespec (str): The optional data file to read

        Returns:
            (FilterData): An instance of this object
        """
        self.tags = dict()
        if filespec:
            self.parse(filespec)

    def parse(self,
              filespec: str,
              ):
        """
        Read in the XLSForm and extract the data we want

        Args:
            filespec (str): The filespec to the XLSForm file

        Returns:
            title (str): The title from the XLSForm Setting sheet
            extract (str): The data extract filename from the XLSForm Survey sheet
        """
        excel_object = pd.ExcelFile(filespec)
        entries = excel_object.parse(sheet_name=[0, 1, 2], index_col=0,
                                     usercols=[0, 1, 2])
        entries = pd.read_excel(filespec, sheet_name=[0, 1, 2])
        title = entries[2]['form_title'].to_list()[0]
        extract = ""
        for entry in entries[0]['type']:
            if str(entry) == 'nan':
                continue
            if entry[:20] == "select_one_from_file":
                extract = entry[21:]
                log.info(f"Got data extract filename: \"{extract}\", title: \"{title}\"")
            else:
                extract = "none"
        total = len(entries[1]['list_name'])
        index = 1
        while index < total:
            key = entries[1]['list_name'][index]
            if key == 'model' or str(key) == "nan":
                index += 1
                continue
            value = entries[1]['name'][index]
            if value == "<text>" or str(value) == "null":
                index += 1
                continue
            if key not in self.tags:
                self.tags[key] = list()
            self.tags[key].append(value)
            index += 1

        # # The yaml config file for the query has a list of columns
        # # to keep in addition to this default set.
        # path = xlsforms_path.replace("xlsforms", "data_models")
        # category = os.path.basename(filespec).replace(".xls", "")
        # file = open(f"{path}/{category}.yaml", "r").read()
        # self.yaml = yaml.load(file, Loader=yaml.Loader)
        # keep = ("name",
        #         "name:en",
        #         "id",
        #         "operator",
        #         "addr:street",
        #         "addr:housenumber",
        #         "osm_id",
        #         "title",
        #         "tags",
        #         "label",
        #         "landuse",
        #         "opening_hours",
        #         "tourism",
        #         )
        # self.keep = list(keep)
        # if 'keep' in self.yaml:
        #     self.keep.extend(self.yaml['keep'])

        return title, extract

    def cleanData(self,
                  data,
                  ):
        """
        Filter out any data not in the data_model

        Args:
            data (bytes): The input data or filespec to the input data file

        Returns:
            (FeatureCollection): The modifed data
        
        """
        if type(data) == str:
            outfile = open(f"new-{data}", "x")
            infile = open(tmpfile, "r")
            indata = geojson.load(infile)
        elif type(data) == bytes:
            indata = eval(data.decode())
        else:
            indata = data
        # these just create noise in the log file
        ignore = (
            "timestamp",
            "version",
            "changeset",
            )
        keep = ('osm_id', 'id', 'version')
        collection = list()
        for feature in indata['features']:
            properties = dict()
            for key, value in feature['properties'].items():
                log.debug(f"{key} = {value}")
                # if key in self.keep:
                if False:
                    if key == 'tags':
                        for k, v in value.items():
                            if k[:4] == "name":
                                properties['title'] = value[k]
                                properties['label'] = value[k]
                            else:
                                properties[k] = v
                    else:
                        if key == 'osm_id':
                            properties['id'] = value
                        else:
                            properties[key] = value
                else:
                    if key in keep:
                        properties[key] = value
                        continue
                    if key in self.tags:
                        if key == "name" or key == 'name:en':
                            properties['title'] = self.tags[key]
                            properties['label'] = self.tags[key]
                        if value in self.tags[key]:
                            properties[key] = value
                        else:
                            if value != "yes":
                                log.warning(f"Value {value} not in the data model!")
                            continue
                    else:
                        if key in ignore:
                            continue
                        log.warning(f"Tag {key} not in the data model!")
                        continue
            newfeature = Feature(geometry=feature['geometry'], properties=properties)
            collection.append(newfeature)
        if type(data) == str:
            geojson.dump(FeatureCollection(collection), outfile)
        return FeatureCollection(collection)

def main():
    """This main function lets this class be run standalone by a bash script"""
    parser = argparse.ArgumentParser(
        description="Convert ODK XML instance file to CSV format"
    )
    parser.add_argument("-v", "--verbose", nargs="?", const="0", help="verbose output")
    parser.add_argument("-i", "--infile", required=True, help="The data extract for ODK Collect")
    parser.add_argument("-x", "--xform",  required=True, help="The XForm for ODK Collect")
    parser.add_argument("-o", "--outfile", default="models.yaml", help="The Yaml file of all tags and values")
    args = parser.parse_args()

    # if verbose, dump to the termina
    if args.verbose is not None:
        log.setLevel(logging.DEBUG)
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(threadName)10s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        log.addHandler(ch)

    xls = FilterData()
    path = xlsforms_path.replace("xlsforms", "data_models")
    models = FilterData()
    data = models.parse(args.xform)
    if args.infile:
        cleaned = models.cleanData(args.infile)
    else:
        if os.path.exists(args.outfile):
            os.remove(args.outfile)
        outfile = open(args.outfile, "w")
        current = None
        tab = "    "
        outfile.write("""
        # Do not edit this file, it is generated by the filter_data.py
        # script from the XForm spreadsheet.
        """)
        for key, value in data.items():
            if key != current:
                outfile.write(f"\n{key}:\n")
                for val in value:
                    outfile.write(f"{tab}- {val}\n")

if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

# Copyright (c) 2022, 2023 Humanitarian OpenStreetMap Team
#
# This program is free software: you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.

import argparse
import logging
import sys
import os
from sys import argv
from osm_fieldwork.osmfile import OsmFile
from geojson import Point, Feature, FeatureCollection, dump, Polygon
import geojson
import psycopg2
from shapely.geometry import shape, Polygon, mapping
import shapely
from shapely import wkt
import xmltodict
from progress.bar import Bar, PixelBar
from progress.spinner import PixelSpinner
from osm_fieldwork.convert import escape
from osm_fieldwork.make_data_extract import PostgresClient, uriParser
from codetiming import Timer
import concurrent.futures
from cpuinfo import get_cpu_info
from time import sleep
from haversine import haversine, Unit
from thefuzz import fuzz, process

# Instantiate logger
log = logging.getLogger(__name__)

# The number of threads is based on the CPU cores
info = get_cpu_info()
cores = info['count']


class OdkMerge(object):
    def __init__(self,
                 source: str,
                 boundary: str = None,
                 ):
        """
        Initialize Input data source

        Args:
            source (str): 
            boundary: str = None

        Returns:
            (OdkMerge): An instance of this object
        """
        self.postgres = list()
        self.source = source
        self.tags = dict()
        # Distance in meters for conflating with postgis
        self.tolerance = 7
        self.data = dict()
        self.analyze = ("name", "amenity", "landuse", "cuisine", "tourism", "leisure")
        # PG: is the same prefix as ogr2ogr
        # "[user[:password]@][netloc][:port][/dbname]"
        if source[0:3] == "PG:":
            uri = uriParser(source[3:])
            # self.source = "underpass" is not support yet
            # Each thread needs it's own connection to postgres to avoid problems.
            for thread in range(0, cores + 1):
                db = PostgresClient(dbhost=uri['dbhost'], dbname=uri['dbname'], dbuser=uri['dbuser'], dbpass=uri['dbpass'])
                self.postgres.append(db)
                if boundary:
                    self.clip(boundary, db)
        else:
            log.info("Opening data file: %s" % source)
            src = open(source, "r")
            self.data = geojson.load(src)
            if boundary:
                self.clip(boundary)

    def clip(self,
             boundary: str,
             db: PostgresClient,
             ):
        """
        Clip a data source by a boundary

        Args:
             boundary (str): The filespec of the project AOI
             db (PostgresClient): A reference to the existing database connection

        Returns:
            (bool): If the region was clipped sucessfully
        """
        remove = list()
        if not boundary:
            return False

        if type(boundary) != dict:
            clip = open(boundary, "r")
            geom = geojson.load(clip)
            if 'features' in geom:
                poly = geom['features'][0]['geometry']
            else:
                poly = geom["geometry"]
        else:
            poly = boundary
        ewkt = shape(poly)

        if len(self.data) > 0:
            # TODO: FMTM produces data extracts the exact size of the boundary
            # polygon, so  we don't need to clip it. In the future though we
            # want this to produce a subset from a larger file.
            for feature in self.data['features']:
                entry = shapely.from_geojson(str(feature))
            #     if not shapely.contains(ewkt, entry):
            #         log.debug(f"CONTAINS {entry}")
            #         del self.data[self.data['features']]
            pass
        else:
            # setup the postgres VIEWs with a dummy SQL query
            sql = f"SELECT COUNT(osm_id) FROM nodes"
            result = db.queryLocal(sql, ewkt)
        return True

    def makeNewFeature(self,
                       attrs: dict = None,
                       tags: dict = None,
                       ):
        """
        Create a new feature with optional data

        Args:
            attrs (dict): All of the attributes and their values
            tags (dict): All of the tags and their values

        Returns:
            (dict): A template feature with no data
        """
        newf = dict()
        if attrs:
            newf['attrs'] = attrs
        else:
            newf['attrs'] = dict()
        if tags:
            newf['tags'] = tags
        else:
            newf['tags'] = dict()
        return newf

    def conflateFile(self,
                     feature: dict,
                     ):
        """
        Conflate a POI against all the features in a GeoJson file

        Args:
            feature (dict): The feature to conflate

        Returns:
            (dict):  The modified feature
        """
        # Most smartphone GPS are 5-10m off most of the time, plus sometimes
        # we're standing in front of an amenity and recording that location
        # instead of in the building.
        gps_accuracy = 10
        # this is the treshold for fuzzy string matching
        match_threshold = 80
        # log.debug(f"conflateFile({feature})")
        hits = False
        data = dict()
        geom = Point((float(feature["attrs"]["lon"]), float(feature["attrs"]["lat"])))
        wkt = shape(geom)
        for existing in self.data['features']:
            id = int(existing['properties']['id'])
            entry = shapely.from_geojson(str(existing))
            if entry.geom_type != 'Point':
                center = shapely.centroid(entry)
            else:
                center = entry
                # dist = shapely.hausdorff_distance(center, wkt)
                # if 'name' in existing['properties']:
                #     print(f"DIST1: {dist}, {existing['properties']['name']}")
            # x = shapely.distance(wkt, entry)
            # haversine reverses the order of lat & lon from what shapely uses. We
            # use this as meters is easier to deal with than cartesian coordinates.
            x1 = (center.coords[0][1], center.coords[0][0])
            x2 = (wkt.coords[0][1], wkt.coords[0][0])
            dist = haversine(x1, x2, unit=Unit.METERS)
            if dist < gps_accuracy:
                # if 'name' in existing['properties']:
                # log.debug(f"DIST2: {dist}")
                # log.debug(f"Got a Hit! {feature['tags']['name']}")
                for key,value in feature['tags'].items():
                    if key in self.analyze:
                        if key in existing['properties']:
                            result = fuzz.ratio(value, existing['properties'][key])
                            if result > match_threshold:
                                # log.debug(f"Matched: {result}: {feature['tags']['name']}")
                                existing['properties']['fixme'] = "Probably a duplicate!"
                                log.debug(f"Got a dup in file!!! {existing['properties']['name'] }")
                                hits = True
                                break
            if hits:
                version = int(existing['properties']['version'])
                # coords = feature['geometry']['coordinates']
                # lat = coords[1]
                # lon = coords[0]
                attrs = {'id': id, 'version': version, 'lat': feature['attrs']['lat'], 'lon': feature['attrs']['lon']}
                tags = existing['properties']
                tags['fixme'] = "Probably a duplicate!"
                # Data extracts for ODK Collect
                del tags['title']
                del tags['label']
                if 'building' in tags:
                    return {'attrs': attrs, 'tags': tags, 'refs': list()}
                return {'attrs': attrs, 'tags': tags}
        return dict()

    def conflateWay(self,
                    feature: dict,
                    dbindex: int,
                    ):
        """
        Conflate a POI against all the ways in a postgres view

        Args:
            feature (dict): The feature to conflate
            dbindex (int): An index into the array of postgres connections

        Returns:
            (dict):  The modified feature
        """
        # log.debug(f"conflateWay({feature})")
        hits = False
        result = list()
        geom = Point((float(feature["attrs"]["lon"]), float(feature["attrs"]["lat"])))
        wkt = shape(geom)
        for key, value in feature['tags'].items():
            if key in self.analyze:
                # Sometimes the duplicate is a polygon, really common for parking lots.
                cleanval = escape(value)
                query = f"SELECT osm_id,tags,version,ST_AsText(ST_Centroid(geom)) FROM ways_view WHERE ST_Distance(geom::geography, ST_GeogFromText(\'SRID=4326;{wkt.wkt}\')) < {self.tolerance} AND levenshtein(tags->>'{key}', '{cleanval}') <= 1"
                # log.debug(query)
                self.postgres[dbindex].dbcursor.execute(query)
                try:
                    result = self.postgres[dbindex].dbcursor.fetchall()
                except:
                    result = list()
                    # log.warning(f"No results at all for {query}")
                if len(result) > 0:
                    hits = True
                    break
        if hits:
            log.debug(f"Got a dup in ways!!! {feature['tags']['name']}")
            # the result is a list from what we specify for SELECT
            version = int(result[0][2]) + 1
            attrs = {'id': int(result[0][0]), 'version': version}
            tags = result[0][1]
            tags[f'old_{key}'] = value
            tags['fixme'] = "Probably a duplicate!"
            geom = mapping(shapely.from_wkt(result[0][3]))
            refs = list()
            # FIXME: iterate through the points and find the existing nodes,
            # which I'm not sure
            # is possible
            # SELECT osm_id,tags,version FROM nodes WHERE ST_Contains(geom, ST_GeomFromText('Point(-105.9918636 38.5360821)'));
            # for i in geom['coordinates'][0]:
            #    print(f"XXXXX: {i}")
            return {'attrs': attrs, 'tags': tags, 'refs': refs}
        return dict()

    def conflateNode(self,
                     feature: dict,
                     dbindex: int,
                     ):
        """
        Conflate a POI against all the nodes in the view

        Args:
            feature (dict): The feature to conflate
            dbindex (int): An index into the array of postgres connections

        Returns:
            (dict):  The modified feature
        """
        # log.debug(f"conflateNode({feature})")
        hits = False
        geom = Point((float(feature["attrs"]["lon"]), float(feature["attrs"]["lat"])))
        wkt = shape(geom)
        result = list()
        ratio = 1
        for key,value in feature['tags'].items():
            if key in self.analyze:
                # print("%s = %s" % (key, value))
                # Use a Geography data type to get the answer in meters, which
                # is easier to deal with than degress of the earth.
                cleanval = escape(value)
                query = f"SELECT osm_id,tags,version,ST_AsEWKT(geom) FROM nodes_view WHERE ST_Distance(geom::geography, ST_GeogFromText(\'SRID=4326;{wkt.wkt}\')) < {self.tolerance} AND levenshtein(tags->>'{key}', '{cleanval}') <= {ratio}"
                # print(query)
                # FIXME: this currently only works with a local database, not underpass yet
                self.postgres[dbindex].dbcursor.execute(query)
                try:
                    result = self.postgres[dbindex].dbcursor.fetchall()
                except:
                    result = list()
                    # log.warning(f"No results at all for {query}")
                if len(result) > 0:
                    hits = True
                    break
        if hits:
            log.debug(f"Got a dup in nodes!!! {feature['tags']}")
            version = int(result[0][2]) + 1
            coords = shapely.from_wkt(result[0][3][10:])
            lat = coords.y
            lon = coords.x
            attrs = {'id': int(result[0][0]), 'version': version, 'lat': lat, 'lon': lon}
            tags = result[0][1]
            tags[f'old_{key}'] = value
            tags['fixme'] = "Probably a duplicate!"
            return {'attrs': attrs, 'tags': tags}
        return dict()

    def conflateById(self,
                     feature: dict,
                     dbindex: int,
                     ):
        """
        Conflate a feature with existing ways using the OSM ID

        Args:
            feature (dict): The feature to conflate
            dbindex (int): An index into the array of postgres connections

        Returns:
            (dict):  The modified feature
        """
        log.debug(f"conflateById({feature})")
        id = int(feature['attrs']['id'])
        if id > 0:
            if self.source[:3] != "PG:":
                sql = f"SELECT osm_id,tags,version,ST_AsText(geom) FROM ways_view WHERE tags->>'id'='{id}'"
                # log.debug(sql)
                self.postgres[0].dbcursor.execute(sql)
                result = self.postgres[0].dbcursor.fetchone()
                if result:
                    version = int(result[0][2]) + 1
                    attrs = {'id': int(result[0][0]), 'version': version}
                    tags = result[0][1]
                    # tags[f'old_{key}'] = value
                    tags['fixme'] = "Probably a duplicate!"
                    geom = mapping(shapely.from_wkt(result[0][3]))
                    return {'attrs': attrs, 'tags': tags}
                else:
                    sql = f"SELECT osm_id,tags,version,ST_AsText(geom) FROM ways_view WHERE tags->>'id'='{id}'"
                    # log.debug(sql)
                    self.postgres[dbindex].dbcursor.execute(sql)
                    result = self.postgres[dbindex].dbcursor.fetchone()
                    if result:
                        version = int(result[0][2]) + 1
                        attrs = {'id': int(result[0][0]), 'version': version}
                        tags = result[0][1]
                        # tags[f'old_{key}'] = value
                        tags['fixme'] = "Probably a duplicate!"
                        geom = mapping(shapely.from_wkt(result[0][3]))
                    return {'attrs': attrs, 'tags': tags, 'refs': refs}
            else:
                for key, value in self.data.items():
                    if key == id:
                        return value
        return dict()

    def cleanFeature(self,
                     feature: dict,
                     ):
        """
        Remove tags that are attributes instead
        Args:
            feature (dict): The feature to clean

        Returns:
            (dict):  The modified feature
        """
            # We only use the version and ID in the attributes
        if 'id' in feature['tags']:
            del feature['tags']['id']
        if 'version' in feature['tags']:
            del feature['tags']['version']
        if 'title' in feature['tags']:
            del feature['tags']['title']
        if 'label' in feature['tags']:
            del feature['tags']['label']
        return feature

    def dump(self):
        """Dump internal data"""
        print(f"Data source is: {self.source}")
        print(f"There are {len(self.data)} existing features")
        # if len(self.versions) > 0:
        #     for k, v in self.original.items():
        #         print(f"{k}(v{self.versions[k]}) = {v}")

    def conflateData(self,
                     odkdata: list,
                     ):
        """
        Conflate all the data. This the primary interfacte for conflation.

        Args:
            odkdata (list): A list of all the entries in the OSM XML input file

        Returns:
            (dict):  The modified features
        """
        timer = Timer(text="conflateData() took {seconds:.0f}s")
        timer.start()
        # Use fuzzy string matching to handle minor issues in the name column,
        # which is often used to match an amenity.
        if len(self.data) == 0:
            self.postgres[0].dbcursor.execute("CREATE EXTENSION IF NOT EXISTS fuzzystrmatch")
        log.debug(f"OdkMerge::conflateData() called! {len(odkdata)} features")

        # A chunk is a group of threads
        chunk = round(len(odkdata) / cores)

        # cycle = range(0, len(odkdata), chunk)

        # Chop the data into a subset for each thread
        newdata = list()
        future = None
        result = None
        index = 0
        with concurrent.futures.ThreadPoolExecutor(max_workers=cores) as executor:
            i = 0
            subset = dict()
            futures = list()
            for key, value in odkdata.items():
                subset[key] = value
                if i == chunk:
                    i = 0
                    result = executor.submit(conflateThread, subset, self, index)
                    index += 1
                    # result.add_done_callback(callback)
                    futures.append(result)
                    subset = dict()
                i += 1
            for future in concurrent.futures.as_completed(futures):
            # # for future in concurrent.futures.wait(futures, return_when='ALL_COMPLETED'):
                log.debug(f"Waiting for thread to complete..")
                # print(f"YYEESS!! {future.result(timeout=10)}")
                newdata.append(future.result(timeout=5))
        timer.stop()
        return newdata
        # return alldata

def conflateThread(features: dict,
                   source: str,
                   dbindex: int,
                   ):
    """
    Conflate a subset of the data

    Args:
        feature (dict): The feature to conflate
        source (str): The data source for conflation, file or database
        dbindex (int): An index into the array of postgres connections

    Returns:
    """
    timer = Timer(text="conflateThread() took {seconds:.0f}s")
    timer.start()
    log.debug(f"conflateThread() called! {len(features)} features")
    merged = list()
    result = dict()
    dups = 0
    # This is brute force, slow but accurate. Process each feature
    # and look for a possible match with existing data.
    for key, value in features.items():
        id = int(value['attrs']['id'])
        # Each of the conflation methods take a single feature
        # as a parameter, and returns a possible match or a zero
        # length dictionary.
        if id > 0:
            # Any feature ID greater than zero is existing data.
            if source.source[:3] != "PG:":
                result = source.conflateFile(value)
            else:
                # Any feature ID less than zero is new data collected
                # using geopoint in the XLSForm.
                result = source.conflateById(value, dbindex)
        elif id < 0:
            if source.source[:3] != "PG:":
                result = source.conflateFile(value)
            else:
                result = source.conflateNode(value, dbindex)
                if len(result) == 0:
                    result = source.conflateWay(value, dbindex)
        if result and len(result) > 0:
            # Merge the tags and attributes together, the OSM data and ODK data.
            # If no match is found, the ODK data is used to create a new feature.
            if 'fixme' in result['tags']:
                dups += 1
                # newf = source.cleanFeature(result)
                attrs = value['attrs'] | result['attrs']
                tags = value['tags'] | result['tags']
                merged.append({'attrs': attrs, 'tags': tags})
            else:
                merged.append(value)
        else:
            merged.append(value)
            # log.error(f"There are no results!")
    timer.stop()
    log.debug(f"Found {dups} duplicates")
    return merged


def main():
    """This main function lets this class be run standalone by a bash script"""
    parser = argparse.ArgumentParser(
        prog="odk_merge.py",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="This program conflates ODK data with existing features from OSM.",
        epilog="""
    This program conflates the output from Osm-Fieldwork after converting
    the JSON file from ODK Central to OSM XML. The data source for existing data can'
    be either the data extract used by the XLSForm, or a postgresql database.

    examples:
        odk_merge.py [OPTIONS] [OSMFILE] [DATAFILE]
        Where OSMFILE is the output from json2osm.py
        And DATAFILE is either the data extract from make_data_extract.py,
        OR refix with "PG:[url]" to use a postgres database
        """,
    )
    parser.add_argument("-v", "--verbose",  action="store_true", help="verbose output")
    parser.add_argument("-o", "--outfile",  help="Output file from the conflation")
    # parser.add_argument("-e", "--extract",  help="OSM data extract created by Osm-Fieldwork")
    # parser.add_argument("-f", "--odkfile",  required=True, help="OSM XML file created by Osm-Field")
    parser.add_argument("-b", "--boundary", help="Boundary polygon to limit the data size")

    args, unknown = parser.parse_known_args()
    osmdata = None
    source = None
    if len(unknown) < 2:
        parser.print_help()
        quit()
    else:
        osmdata = unknown[0]
        source = unknown[1]

    # if verbose, dump to the terminal.
    if args.verbose:
        log.setLevel(logging.DEBUG)
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(threadName)10s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        log.addHandler(ch)

    if args.outfile:
        outfile = args.outfile
    else:
        outfile = os.path.basename(osmdata.replace('.osm', '-foo.osm'))

    # This is the existing OSM data, a database or a file
    if args.boundary:
        extract = OdkMerge(source, args.boundary)
    else:
        extract = OdkMerge(source)

    if extract:
        odkf = OsmFile(outfile) # output file
        osm = odkf.loadFile(osmdata) # input file
        #odkf.dump()
    else:
        log.error("No ODK data source specified!")
        parser.print_help()
        quit()

    # This returns a list of lists of dictionaries. Each thread returns
    # a list of the features, and len(data) is thre number of CPU cores.
    data = extract.conflateData(osm)
    out = list()
    #print(data)
    for entry in data:
        # if 'refs' in feature or 'building' in feature['tags']:
        for feature in entry:
            if 'refs' in feature:
                feature['refs'] = list()
                out.append(odkf.createWay(feature, True))
            else:
                out.append(odkf.createNode(feature, True))

    # out = ""
    # for id, feature in osm.items():
    #     result = extract.conflateFile(feature)
    #     if len(result) > 0:
    #         node = odkf.featureToNode(result)
    #     else:
    #         node = feature
    #     out += odkf.createNode(node, True)
    odkf.write(out)
    log.info(f"Wrote {outfile}")

if __name__ == "__main__":
    """This is just a hook so this file can be run standlone during development."""
    main()



#!/usr/bin/python3

# Copyright (c) 2021, 2022, 2023 Humanitarian OpenStreetMap Team
#
# This file is part of Osm-Fieldwork.
#
#     Osm-Fieldwork is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Underpass is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Osm-Fieldwork.  If not, see <https:#www.gnu.org/licenses/>.
#

import os
import sys
import argparse
from osm_fieldwork.xlsforms import xlsforms_path
from osm_fieldwork.convert import Convert

# find the path of root tests dir
rootdir = os.path.dirname(os.path.abspath(__file__))

parser = argparse.ArgumentParser(
    description="Read and convert a JSON file from ODK Central"
)
parser.add_argument("--infile", default=f"{rootdir}/testdata/testcamps.json", help="The JSON input file")
args = parser.parse_args()

path = xlsforms_path.replace("/xlsforms", "")
csv = Convert(f"{path}/xforms.yaml")

def test_keywords():
    """Convert a feature"""
    hits = 0
    if csv.convertData('fee'):
        hits += 1 
    if csv.convertData('sac_scale') is False:
        hits += 1 
    assert hits == 2

def test_convert_tag():
    """Test tag conversion"""
    hits = 0
    # Test a tag that gets converted
    if csv.convertTag("altitude") == 'ele':
        hits += 1
    # Test a tag that doesn't get converted
    if csv.convertTag("foobar") == 'foobar':
        hits += 1
    assert hits == 2

def test_single_value():
    """Test tag value conversion"""
    hits = 0
    # Test a value that gets converted
    if csv.convertValue("building:floor", 'wood'):
        hits += 1
    assert hits == 1

def test_sub_value():
    """Test tag value conversion"""
    hits = 0
    # Test a value that gets converted
    val = csv.convertValue("cemetery_services", 'cemetery')
    if len(val) == 1 and val[0]['amenity'] == 'grave_yard':
        hits += 1
    assert hits == 1

def test_multiple_value():
    """Test tag value conversion"""
    hits = 0
    # Test a value that gets converted
    vals = csv.convertValue('amenity', 'coffee')
    if len(vals) == 2 and vals[0]['amenity'] == 'cafe' and vals[1]['cuisine'] == 'coffee_shop':
        hits += 1
    assert hits == 1

# Run standalone for easier debugging when not under pytest
if __name__ == "__main__":
    test_keywords()
    test_convert_tag()
    test_single_value()
    test_sub_value()
    test_multiple_value()
